import logging

# Setup logging
logger = logging.getLogger("pdf-renamer")
logger.setLevel(level=logging.INFO)
if not logger.handlers:
    formatter = logging.Formatter("[pdf-renamer]: %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
logger.propagate = False

from .config import config
config.ReadParamsINIfile()

config.set('verbose',config.get('verbose')) #This is a quick and dirty way (to improve in the future) to make sure that the verbosity of the pdf2doi logger is properly set according
                                            #to the current value of config.get('verbose') (see config.py file for details)

from .main import rename,build_filename
from .filename_creators import *




'''
This module contains functions that are used to add or remove keys to the registry of Windows, for the purposes
of accessing pdf-renamer by right-clicking on a pdf file or a directory
'''

import logging
from sys import executable as python_path
from os import path
import os

if os.name == 'nt':
    import winreg as reg

logger = logging.getLogger("pdf-renamer")



def delete_sub_key(key0, current_key, arch_key=0):
    #Code inpsired by Orsiris de Jong's solution https://stackoverflow.com/questions/38205784/python-how-to-delete-registry-key-and-subkeys-from-hklm-getting-error-5
    open_key = reg.OpenKey(key0, current_key, 0, reg.KEY_ALL_ACCESS | arch_key)
    info_key = reg.QueryInfoKey(open_key)
    for x in range(0, info_key[0]):
        # NOTE:: This code is to delete the key and all sub_keys.
        # If you just want to walk through them, then
        # you should pass x to EnumKey. sub_key = reg.EnumKey(open_key, x)
        # Deleting the sub_key will change the sub_key count used by EnumKey.
        # We must always pass 0 to EnumKey so we
        # always get back the new first sub_key.
        sub_key =reg.EnumKey(open_key, 0)
        try:
            reg.DeleteKey(open_key, sub_key)
            logger.info("Removed %s\\%s " % (current_key, sub_key))
        except OSError:
            delete_sub_key(key0, "\\".join([current_key,sub_key]), arch_key)
            # No extra delete here since each call
            # to delete_sub_key will try to delete itself when its empty.

    reg.DeleteKey(open_key, "")
    open_key.Close()
    logger.info("Removed %s" % current_key)
    return

def install_right_click():
    if not(os.name == 'nt'):
        logger.error(f'This functionality is currently implemented only for Windows.')
        return
    python_folder = path.dirname(python_path)
    if python_folder[-7:].lower() == 'scripts': #This typically happens when python is installed in a virtual environment
        path_exec = python_folder + "\pdfrenamer.exe"
    else:
        path_exec = python_folder + "\scripts\pdfrenamer.exe"
    logger.info(f'Adding pdf-renamer to the right-click context menu by adding keys to the system register...')
    try:

        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'Directory\shell\pdfrenamer')
        reg.SetValueEx(key, 'MUIVerb', 0, reg.REG_SZ, 'pdf-renamer')
        reg.SetValueEx(key, 'subcommands', 0, reg.REG_SZ, '')
        reg.CloseKey(key)
        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'Directory\shell\pdfrenamer\shell')
        reg.CloseKey(key)

        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'Directory\shell\pdfrenamer\shell\pdfrenamer_rename')
        reg.SetValue(key, '', reg.REG_SZ, 'Rename all PDF files in this folder (use default settings)...')
        reg.CloseKey(key)
        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'Directory\shell\pdfrenamer\shell\pdfrenamer_rename\command')
        reg.SetValue(key, '', reg.REG_SZ, path_exec + " \"%1\"")
        reg.CloseKey(key)


        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'SystemFileAssociations\.pdf\shell\pdfrenamer')
        reg.SetValueEx(key, 'MUIVerb', 0, reg.REG_SZ, 'pdf-renamer')
        reg.SetValueEx(key, 'subcommands', 0, reg.REG_SZ, '')
        reg.CloseKey(key)
        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'SystemFileAssociations\.pdf\shell\pdfrenamer\shell')
        reg.CloseKey(key)

        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'SystemFileAssociations\.pdf\shell\pdfrenamer\shell\pdfrenamer_rename')
        reg.SetValue(key, '', reg.REG_SZ, 'Rename this file (use default settings)...')
        reg.CloseKey(key)
        key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, 'SystemFileAssociations\.pdf\shell\pdfrenamer\shell\pdfrenamer_rename\command')
        reg.SetValue(key, '', reg.REG_SZ, path_exec + " \"%1\"")
        reg.CloseKey(key)

        logger.info(f'All necessary keys were added to the system register.')
    except Exception as e:
        logger.error(e)
        logger.error(f'A problem occurred when trying to add pdf-renamer to the right-click context menu.\nNOTE: this functionality is only available in Windows, and it has to be installed from a terminal with administrator rights.')
 
def uninstall_right_click():
    if not(os.name == 'nt'):
        logger.error(f'This functionality is currently implemented only for Windows.')
        return
    logger.info(f'Removing all keys associated to pdf-renamer from the system register...')
    try:
        delete_sub_key(reg.HKEY_CLASSES_ROOT, "SystemFileAssociations\.pdf\shell\pdfrenamer")
        delete_sub_key(reg.HKEY_CLASSES_ROOT, "Directory\shell\pdfrenamer")
        logger.info(f'All keys were removed.')
    except Exception as e:
        logger.error(e)
        logger.error(f'A problem occurred when trying to remove keys from the system registry.\nNOTE: this command needs to be executed from a terminal with administrator rights.')
 



import argparse
import logging
import bibtexparser
import pathlib
import os
#from os import path, listdir, scandir
import pdf2bib
import pdf2doi
#import itertools
#import pkgutil
import pdfrenamer.config as config
from pdfrenamer.filename_creators import build_filename, AllowedTags, check_format_is_valid
import traceback
import sys

#We tell pdf2doi and pdf2bib to use the same value of save_identifier_metadata specified in the settings of pdf-renamer. When pdf-renamer is called via command line, 
#the value of save_identifier_metadata (for both pdf2doi, pdf2bib and pdf-renamer) might get changed
pdf2doi.config.set('save_identifier_metadata',config.get('add_metadata')) 
pdf2bib.config.set('save_identifier_metadata',config.get('add_metadata'))  

logger = logging.getLogger("pdf-renamer")

def rename(target, format=None, tags=None):
    '''
    This is the main routine of the script. When the library is used as a command-line tool (via the entry-point "pdfrenamer") the input arguments
    are collected, validated and sent to this function (see the function main () below). 
    The function tries to rename the pdf file whose path is specified in the input argument target with the format specified in the input 
    argument format. The info of the paper (title, authors, etc.) are obtained via the library pdf2bib (which in turns uses pdf2doi). 
    If the input argument target is the path of a folder, the function is applied to each pdf file contained in the folder
    (by calling again the same function). If the global settingcheck_subfolders is set to True, it also renames pdf files in all subfolders (recursively).

    Parameters
    ----------
    target : string
        Relative or absolute path of the target .pdf file or directory
    Returns
    -------
    results, dictionary or list of dictionaries (or None if an error occured)
        The output is a single dictionary if target is a file, or a list of dictionaries if target is a directory, 
        each element of the list describing one file. Each dictionary has the following keys
        result['path_original']     = path of the pdf file (with the original filename)
        result['path_new']          = path of the pdf file, with the new filename, or None if it was not possible to generate a new filename
        result['identifier']        = DOI or other identifier (or None if nothing is found)
        result['identifier_type']   = String specifying the type of identifier (e.g. 'doi' or 'arxiv')
        result['validation_info']   = Additional info on the paper. If config.get('webvalidation') = True, then result['validation_info']
                                      will typically contain raw bibtex data for this paper. Otherwise it will just contain True 
        result['path']              = Path of the pdf file
        result['method']            = Method used by pdf2doi to find the identifier
        result['metadata']          = Dictionary containing bibtex info
        result['bibtex']            = A string containing a valid bibtex entry

    '''
    
    # Setup logging
    logger = logging.getLogger("pdf-renamer")

    if not format: format = config.get('format')
    
    #Make some sanity check on the format, and extract tags
    if not tags:    #If tags is a valid variable, it means the format was already checked earlier (i.e. this call to the function rename was generated by
                    #an earlier call to the function rename. By not checking again the format, we save time
        tags = check_format_is_valid(format)
        if tags == None: #if the function check_format_is_valid has returned, then the format is not valid and the function terminates
            return None

    #Check if path is valid
    if not(os.path.exists(target)):
        logger.error(f"{target} is not a valid path to a file or a directory.")
        return
    
    #Check if target is a directory
        # If yes, we look for all the .pdf files inside it, and for each of them
        # we call again this function by passing the file path as target.
        #
        # Moreover, if config.get('check_subfolders')==True, for each subfolder in the directory we call again this function
        # by passsing the subfolder as target

    if  os.path.isdir(target):
        logger.info(f"Looking for pdf files and subfolders in the folder {target}...")
        if not(target.endswith(os.path.sep)): #Make sure the path ends with "\" or "/" (according to the OS)
                target = target + os.path.sep

        #We build a list of all the pdf files in this folder, and of all subfolders
        pdf_files = [f for f in os.listdir(target) if (f.lower()).endswith('.pdf')]
        subfolders = [ f.path for f in os.scandir(target) if f.is_dir() ]

        numb_files = len(pdf_files)
        if numb_files == 0:
            logger.error("No pdf file found in this folder.")
        else:
            logger.info(f"Found {numb_files} pdf file(s).")

            files_processed = [] #For each pdf file in the target folder we will store a dictionary inside this list
            for f in pdf_files:
                logger.info(f"................") 
                file = target + f
                #We call again this same function but this time targeting the single file
                result = rename(file, format=format, tags=tags)
                files_processed.append(result)
            logger.info("................") 

        #If there are subfolders, and if config.get('check_subfolders')==True, we call gain this function for each subfolder
        numb_subfolders = len(subfolders)
        if numb_subfolders:
            logger.info(f"Found {numb_subfolders} subfolder(s)")
            if config.get('check_subfolders')==True :
                logger.info("Exploring subfolders...") 
                for subfolder in subfolders:
                    result = rename(subfolder, format=format,tags=tags)
                    files_processed.extend(result)
            else:
                logger.info("The subfolder(s) will not be scanned because the parameter check_subfolders is set to False."+
                            " When using this script from command line, use the option -sf to explore also subfolders.") 
            logger.info("................") 
        return files_processed
    
    #If target is not a directory, we check that it is an existing file and that it ends with .pdf
    else:
        filename = target
        logger.info(f"File: {filename}")  
        if not os.path.exists(filename):
            logger.error(f"'{filename}' is not a valid file or directory.")
            return None    
        if not (filename.lower()).endswith('.pdf'):
            logger.error("The file must have .pdf extension.")
            return None

        if check_if_file_was_already_renamed_with_same_format(filename,format)==True and config.get('force_rename')==False:
            logger.info(f"Based on the pdf metadata, this file has been already renamed by pdf-renamer, and with the same filename format. " + 
                        "Nothing will be done. To overrule this behavior add the command -fr to the pdf-renamer invokation.")
            result = dict()
            result['identifier'] = 'previously_found'
            result['path_original'] = filename
            result['path_new'] = filename
            return result
        
        #We use the pdf2bib library to retrieve info of this file
        logger.info(f"Calling the pdf2bib library to retrieve the bibtex info of this file.")
        try:
            result = pdf2bib.pdf2bib_singlefile(filename)
            result['path_original'] = filename

            #if pdf2bib was able to find an identifer, and thus to retrieve the bibtex data, we use them to rename the file
            if result['metadata'] and result['identifier']:
                logger.info(f"Found bibtex data and an identifier for this file: {result['identifier']} ({result['identifier_type']}).")
                metadata = result['metadata'].copy()
                metadata_string = "\n\t"+"\n\t".join([f"{key} = \"{metadata[key]}\"" for key in metadata.keys()] ) 
                logger.info("Found the following data:" + metadata_string)

                #Generate the new name by calling the function build_filename
                NewName = build_filename(metadata, format, tags)
                ext = os.path.splitext(filename)[-1].lower() #Extract the file extension from the old file name
                directory = pathlib.Path(filename).parent
                NewPath = str(directory) + os.path.sep + NewName
                NewPathWithExt = NewPath + ext
                logger.info(f"The new file name is {NewPathWithExt}")
                if (filename==NewPathWithExt):
                    logger.info("The new file name is identical to the old one. Nothing will be changed")
                    pdf2doi.add_metadata(filename,'/pdfrenamer_nameformat',format)
                    result['path_new'] = NewPathWithExt
                else:
                    try:
                        NewPathWithExt_renamed = rename_file(filename,NewPath,ext) 
                        logger.info(f"File renamed correctly.")
                        if config.get('add_metadata') == True:
                            pdf2doi.add_metadata(NewPathWithExt_renamed ,'/pdfrenamer_nameformat',format)
                        if not (NewPathWithExt == NewPathWithExt_renamed):
                            logger.info(f"(Note: Another file with the same name was already present in the same folder, so a numerical index was added at the end).")
                        result['path_new'] = NewPathWithExt_renamed
                    except Exception as e: 
                        logger.error('Some error occured while trying to rename this file: \n '+ str(e))
                        result['path_new'] = None
            else:
                logger.info("The pdf2doi library was not able to find an identifier for this pdf file.")
                result['path_new'] = None
        except Exception as e: 
            print(traceback.format_exc())
            # or
            print(sys.exc_info()[2])
            logger.error('Some unexpected error occured while using pdf2bib to process this file: \n '+ str(e))
            result['path_new'] = None

        return result 

def rename_file(old_path,new_path,ext):
    #It attempts to rename the file in old_path with the new name contained in new_path. 
    #If another file with the same name specified by new_path already exists in the same folder, it adds an 
    #incremental number (e.g. "filename.pdf" becomes "filename (2).pdf")

    if not os.path.exists(old_path):
        raise ValueError(f"The file {old_path} does not exist")
    i=1
    while True:
        New_path = new_path + (f" ({i})" if i>1 else "") + ext
        if os.path.exists(New_path):
            i = i+1
            continue
        else:
            os.rename(old_path,New_path)
            return New_path

def check_if_file_was_already_renamed_with_same_format(filename,format):
    flag = False
    try:
        with open(filename, 'rb') as f:
            infos = pdf2doi.get_pdf_info(f)
            if '/pdfrenamer_nameformat' in infos.keys():
                if infos['/pdfrenamer_nameformat'] == format:
                    flag = True
            return flag
    except TypeError:
        logger.exception("File processing error")
        return None
    except Exception as e:
        logger.exception(f"File processing error: {e}")
        return None
    
def add_abbreviations(path_abbreviation_file):
    #Adds the content of the text file specified by the path path_abbreviation_file at the beginning of the file UserDefinedAbbreviations.txt
    if not(os.path.exists(path_abbreviation_file)):
        logger.error(f"{path_abbreviation_file} is not a valid path to a file.")
        return

    logger.info(f"Loading the file {path_abbreviation_file}...")
    try:
        with open(path_abbreviation_file, 'r') as new_abbreviation_file:
            new_abbreviation = new_abbreviation_file.read()
    except Exception as e: 
        logger.error('Some error occured while loading this file: \n '+ str(e))
        return

    logger.info(f"Adding the content of the file {path_abbreviation_file} to the user-specified journal abbreviations...")

    try:
        path_current_directory = os.path.dirname(__file__)
        path_UserDefinedAbbreviations = os.path.join(path_current_directory, 'UserDefinedAbbreviations.txt')
        with open(path_UserDefinedAbbreviations, 'r') as UserDefinedAbbreviations_oldfile:
            UserDefinedAbbreviations_old = UserDefinedAbbreviations_oldfile.read()
        with open(path_UserDefinedAbbreviations, 'w') as UserDefinedAbbreviations_newfile:
            UserDefinedAbbreviations_newfile.write( new_abbreviation )
            UserDefinedAbbreviations_newfile.write('\n')
            UserDefinedAbbreviations_newfile.write( UserDefinedAbbreviations_old )
    except Exception as e: 
        logger.error('Some error occured: \n '+ str(e))
        return

    logger.info(f"The new journal abbreviations were correctly added.")


def main():
    parser = argparse.ArgumentParser( 
                                    description = "Automatically renames pdf files of scientific publications by retrieving their identifiers (e.g. DOI or arxiv ID) and looking up their bibtex infos.",
                                    epilog = "",
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
                        "path",
                        help = "Relative path of the pdf file or of a folder.",
                        metavar = "path",
                        nargs = '*')
    parser.add_argument("-s",
                        "--decrease_verbose",
                        help="Decrease verbosity. By default (i.e. when not using -s), all steps performed by pdf-renamer, pdf2dbib and pdf2doi are documented.",
                        action="store_true")
    parser.add_argument("-read_only",
                    "--read_only",
                    help="By default, pdf-renamer and pdf2doi store some information the metadata of the pdf file in order to speed up subsequent processing. By using this additional option, no metadata is ever added.",
                    action="store_true")
    parser.add_argument('-f', 
                        help=f"Format of the new filename. Default = \"{config.get('format')}\".\n"+
                        "Valid tags:\n"+
                        "\n".join([key+val for key,val in AllowedTags.items()]),
                        action="store", dest="format", type=str, default=config.get('format'))
    parser.add_argument("-sf",
                        "--sub_folders",
                        help=f"Rename also pdf files contained in subfolders of target folder. Default = \"{config.get('check_subfolders')}\".",
                        action="store_true")
    parser.add_argument('-max_length_authors', 
                        help=f"Sets the maximum length of any string related to authors (default={str(config.get('max_length_authors'))}).",
                        action="store", dest="max_length_authors", type=int, default=config.get('max_length_authors'))
    parser.add_argument('-max_length_filename', 
                        help=f"Sets the maximum length of any generated filename. Any filename longer than this will be truncated (default={str(config.get('max_length_filename'))}).",
                        action="store", dest="max_length_filename", type=int, default=config.get('max_length_filename'))
    parser.add_argument('-max_words_title', 
                        help=f"Sets the maximum number of words from the paper title to use for the filename (default={str(config.get('max_words_title'))}).",
                        action="store", dest="max_words_title", type=int, default=config.get('max_words_title'))
    parser.add_argument('-case', 
                        help=f"Possible values are 'camel', 'snake', 'kebab', 'none' (default={str(config.get('case'))}). \n"+
                        "If different from 'none', converts each tag string into either 'camel' (e.g., LoremIpsumDolorSitAmet), 'snake' (e.g., Lorem_ipsum_dolor_sit_amet), or 'kebab' case (e.g., Lorem-ipsum-dolor-sit-amet). \n"+
                        "Note: this will not affect any punctuation symbol or space contained in the filename format by the user.",
                        action="store", dest="case", type=str, default=config.get('case'))
    parser.add_argument(
                        "-add_abbreviation_file",
                        help="The content of the text file specified by PATH_ABBREVIATION_FILE will be added to the user list of journal abbreviations.\n"+
                        "Each row of the text file must have the format \'FULL NAME = ABBREVIATION\'.",
                        action="store", dest="path_abbreviation_file", type=str)
    parser.add_argument(
                        "-fr",
                        "--force_rename",
                        help=f"By default, whenever pdf-renamer renames a pdf file by using a certain filename format, it also stores the format string into a tag of the pdf file.\n"+
                        "In this way if pdf-renamer comes across that same file later, and the current filename format is the same as the one stored in the pdf file tag,\n"+
                        "the file is ignored. By using this command, this behavior is overruled: pdf-renamer will always rename each file it comes across.",
                        action="store_true")
    parser.add_argument(
                        "-sd",
                        "--set_default",
                        help=f"By adding this command, any value specified (in this same command) for the filename format (-f),\n"+
                        "max length of author string (-max_length_authors), max length of filename string (-max_length_filename),\n"+
                        "max number of title words (-max_words_title), and case (-case) will be also stored as default value(s) for the future.",
                        action="store_true")
    parser.add_argument("-install--right--click",
                        dest="install_right_click",
                        action="store_true",
                        help=f"Add a shortcut to pdf-renamer in the right-click context menu of Windows. You can rename a single pdf file (or all pdf files in a folder) by just right clicking on it!\n"+
                       "NOTE: this feature is only available on Windows.")
    parser.add_argument("-uninstall--right--click",
                        dest="uninstall_right_click",
                        action="store_true",
                        help="Uninstall the right-click context menu functionalities. NOTE: this feature is only available on Windows.")

    
    args = parser.parse_args()

    # Setup logging
    config.set('verbose',not(args.decrease_verbose)) #store the desired verbose level in the global config of pdf-renamer. This will also automatically update the pdf2bib and pdf2doi logger level.
    logger = logging.getLogger("pdf-renamer")

    #If the command -install--right--click was specified, it sets the right keys in the system registry
    if args.install_right_click:
        config.set('verbose',True)
        import pdfrenamer.utils_registry as utils_registry
        utils_registry.install_right_click()
        return
    if args.uninstall_right_click:
        config.set('verbose',True)
        import pdfrenamer.utils_registry as utils_registry
        utils_registry.uninstall_right_click()
        return

    if args.path_abbreviation_file:
        add_abbreviations(args.path_abbreviation_file)
        return

    if (check_format_is_valid(args.format)):
        config.set('format' , args.format)

    if (isinstance(args.max_length_authors,int) and args.max_length_authors>0):
        config.set('max_length_authors' , args.max_length_authors)
    else:
        logger.error(f"The specified value for max_length_authors is not valid.")

    if (isinstance(args.max_length_filename,int) and args.max_length_filename>0):
        config.set('max_length_filename' , args.max_length_filename)
    else:
        logger.error(f"The specified value for max_length_filename is not valid.")
    config.set('check_subfolders' , args.sub_folders)

    if (isinstance(args.max_words_title,int) and args.max_words_title>0):
        config.set('max_words_title' , args.max_words_title)
    else:
        logger.error(f"The specified value for max_words_title is not valid.")

    if (isinstance(args.case,str) and (args.case in ['camel','snake','kebab','none'])):
        config.set('case' , args.case)
    else:
        logger.error(f"The specified value for case is not valid.")
        return

    config.set('check_subfolders' , args.sub_folders)
    config.set('force_rename' , args.force_rename)

    if args.set_default:
        logger.info("Storing the settings specified by the user (if any is valid) as default values...")
        config.WriteParamsINIfile()
        logger.info("Done.")

    ## The following block of code (until ##END) is required to make sure that 'path' is considered a required parameter, except for the case when
    ## -install--right--click or -uninstall--right--click are used, or when the user is setting default values for some of the parameters
    if isinstance(args.path,list):
        if len(args.path)>0:
            target = args.path[0]
        else:
            target = ""
    else:
        target = args.path
    if target == "" and not (args.set_default):
        print("pdfrenamer: error: the following arguments are required: path. Type \'pdfrenamer --h\' for a list of commands.")
    if target == "": #This occurs either if the user forgot to add a target, or if the user used the -sd command to set default values
        return
    ## END
    
    config.set('add_metadata', not (args.read_only))
    pdf2doi.config.set('save_identifier_metadata',config.get('add_metadata')) 
    pdf2bib.config.set('save_identifier_metadata',config.get('add_metadata')) 

    if(args.decrease_verbose==True):
        print(f"(All intermediate output will be suppressed. To see additional output, do not use the command -s)")
    results = rename(target=target)

    if results==None:  #This typically happens when target is neither a valid file nor a valid directory. In this case we stop
        return         #the script execution here. Proper error messages were raised by the rename function

    if  os.path.isdir(target):
        target = os.path.join(target, '') #This makes sure that, if target is a path to a directory, it has the ending "/" or "\"
    MainPath = os.path.dirname(target) #Extract the path of target. If target is a directory, then MainPath = target

    from colorama import init,Fore, Back, Style
    init(autoreset=True)
    print(Fore.RED + "Summaries of changes done:")

    if not isinstance(results,list):
        results = [results]
    
    counter = 0
    counter_identifier_notfound = 0

    for result in results:
        if result and result['identifier'] and result['path_new']:
            if not(result['path_original']==result['path_new']):
                print(Fore.YELLOW + f"{os.path.relpath(result['path_original'],MainPath)}")
                print(Fore.MAGENTA + f"---> {os.path.relpath(result['path_new'],MainPath)}")
                counter = counter + 1
        elif not (result['identifier']): 
            counter_identifier_notfound = counter_identifier_notfound + 1

    if counter==0:
        print("No file has been renamed.")
    else:
        print(f"{counter} file" + ("s have " if counter>1 else " has ") + "been renamed.")

    if counter_identifier_notfound > 0:
        print(Fore.RED +"The following pdf files could not be renamed because it was not possile to automatically find " +
              "the publication identifier (DOI or arXiv ID). Try to manually add a valid identifier to each file via " +
              "the command \"pdf2doi 'filename.pdf' -id 'valid_identifier'\" and then run again pdf-renamer.")  
        for result in results:
            if not(result['identifier']):
                print(f"{result['path_original']}")
    return

if __name__ == '__main__':
    main()



''' 
This module contains several functions and variables that are used to generate
a valid filename, based on the chosen format and the available infos
'''

import re
import pkgutil
import pdf2bib
import pdfrenamer.config as config
import logging
import unidecode
logger = logging.getLogger("pdf-renamer")

AllowedTags = {"{YYYY}":" \t\t=\t Year of publication",
                "{MM}":" \t\t=\t Month of publication (in digits)",
                "{DD}":" \t\t=\t Day of publication (in digits)",
                "{J}":" \t\t=\t Full name of Journal",
                "{Jabbr}":" \t=\t Abbreviated name of Journal, if any available (otherwise full name is used)",
                "{Aall}":" \t\t=\t Last name of all authors (separated by comma)",
                "{Aetal}":" \t=\t Last name of the first author, add \'et al.\' if more authors are present",
                "{A3etal}":" \t=\t Last name of the first three authors (separated by comma), add \'et al.\' if more authors are present",
                "{aAall}":" \t=\t First initial and last name of all authors (separated by comma)",
                "{aAetal}":" \t=\t First initial and last name of the first author, add \'et al.\' if more authors are present",
                "{aA3etal}":" \t=\t First initial and last name of the first three authors (separated by comma), add \'et al.\' if more authors are present",
                "{T}":" \t\t=\t Title"
                #"{Tcamel}":" \t=\t Title in camel case (e.g., LoremIpsumDolorSitAmet)",
                #"{Tsnake}":" \t=\t Title in snake case (e.g., Lorem_ipsum_dolor_sit_amet)",
                #"{Tkebab}":" \t=\t Title in kebab case (e.g., Lorem-ipsum-dolor-sit-amet)"
                }

valid_months = {'jan':'01','january':'01',
                'feb':'02','february':'02',
                'mar':'03','march':'03',
                'apr':'04','april':'04',
                'may':'05','may':'05',
                'jun':'06','june':'06',
                'jul':'07','july':'07',
                'aug':'08','august':'08',
                'sep':'09','september':'09',
                'oct':'10','october':'10',
                'nov':'11','november':'11',
                'dec':'12','december':'12'}

def to_camel(s):
  s = re.sub(r"([-_,\.\s]+)", " ", s).title().strip().replace(" ", "")
  return ''.join(s)

def to_snake(s):
  s = re.sub(r"([-_,\.\s]+)", " ", s).strip().replace(" ", "_")
  return ''.join(s)

def to_kebab(s):
  s = re.sub(r"([-_,\.\s]+)", " ", s).strip().replace(" ", "-")
  return ''.join(s)


string_transformations = {
                    "camel" : to_camel,
                    "snake" : to_snake,
                    "kebab" : to_kebab,
                    }

def month_to_number(month_string):
    return valid_months.get(month_string,"00")

def validate_journal(journal):
    return journal

def is_valid_integer(string,number_digits):
    if not isinstance(string,str):
        string = str(string)
    return (string.isnumeric() and len(string)==number_digits)

def sanitize(string):
    #Given a string in input, it first removes all possible latex codes, and then removes any residual character which would not be 
    #allowed in a file name

    #Step 1. Replace common latex symbols
    replace ={       
                '-'  : ['{\\textendash}','{\\textemdash}'] ,
                '_'  : ['{\\textunderscore}'],
                ' '  : ["{\\textasteriskcentered}","{\\textgreater}","{\\textless}"],
                ''  : ["\n"],
                '{'  : ["{\\textbraceleft}"],
                '}'  : ["{\\textbraceright}"],
                "\'" : ["{\\textquotesingle}","{\\textquotedblleft}","{\\textquotedblright}","{\\textquoteleft}","{\\textquoteright}"]
                }
    for i, j in replace.items():
        for k in j:
            string = string.replace(k, i)
    string = re.sub(r"{\\hspace{.*}}", "",string)

    #Step 2. Find all substrings in the format {\string1{string2}} (e.g. {\`{u}}) and replace them by string2
    #We use the function remove_latex_codes defined in the pdf2bib package
    string = pdf2bib.remove_latex_codes(string)

    string = unidecode.unidecode(string)    

    #Step 3. #Check that the string is only made out of ascii characters (i.e. no accents, tildes, etc.)
    string = unidecode.unidecode(string) 
    
    #Step 4. Remove any residual special character
    invalid = "<>\"/\|*{}'?:"
    for char in invalid:
        string = string.replace(char, '')
        
    #Step 5. If there are multiple spaces, replace them with only one
    string = re.sub(' +',' ',string)

    return string

def find_abbreviation_journal(journal_name):
    """
    Find a journal abbreviation for a given journal name.

    Parameters
    ----------
    journal_name : string
        Name of the Journal.

    Returns
    -------
    string
        The abbreviation of the journal if any is found, or None if no abbraviation is found

    """
    to_search = sanitize( (journal_name.strip() + " = ").lower() )

    data = pkgutil.get_data(__name__, "UserDefinedAbbreviations.txt").decode('utf8')
    for line in data.splitlines():
        if (line.lower()).startswith(to_search):
            return line[len(to_search):].rstrip()

    data = pkgutil.get_data(__name__, "StandardAbbreviations.txt").decode('utf8')
    for line in data.splitlines():
        if (line.lower()).startswith(to_search):
            return line[len(to_search):].rstrip()
    return None

def find_tags_in_format(format):
    #Given the input string 'format', it creates a list of all the tags "{str}" 
    #contained in it, by looking for stuff wrapped between { and }
    # Example, tags = ['{YYYY}', '{MM}', '{J}', '{A}', '{T}']
    tags = re.findall(r'\{.*?\}', format)       
    return tags

def check_format_is_valid(format):
    #Check if the input string 'format' is a valid format for a filename
    #Returns None if it is not a valid format, or a list of the tags contained in format otherwise
    if not format or (not isinstance(format,str)):
        logger.error(f"The specified format is not a valid string.")
        return None
    tags = find_tags_in_format(format)
    if not tags:
        logger.error(f"The specified format does not contain any tag. Tags must be delimited by {{ and }}.")
        return None
    for tag in tags:
        if not tag in AllowedTags:
            logger.error(f"The specified format contains \"{tag}\", which is not a valid tag.")
            logger.error(f"The valid tags are: " + ",".join(AllowedTags))
            return None
    return tags


def build_filename(infos,   format = None, tags=None):
    '''
    It generates a filename based on the metadata contained in the input dictionary 'infos', using the format specified 
    in the input string 'format'. The tags contained in format have been already identified, and passed to this function 
    in the input dictionary 'tags'. 
    '''
    if not format: format = config.get('format')
    rep_dict =  dict.fromkeys(tags) #Initialize a dictionary with keys equal to the elements of the list tags, and all the values set to None

    for key,value in infos.items():
        if type(value) in (float,int):
            infos[key] = str(value)

    #Now we look in the keys of the 'rep_dict' dictionary and populate the values of the dictionary
    #by using the information contained in the 'infos' dictionary
   
    if '{YYYY}' in rep_dict.keys():
        rep_dict['{YYYY}'] = infos['year'] if ('year' in infos and is_valid_integer(infos['year'],4)) else '0000'

    if '{MM}'in rep_dict.keys():
        rep_dict['{MM}'] = '00'
        if ('month' in infos):
            if is_valid_integer(infos['month'],2):
                rep_dict['{MM}'] = infos['month']
            elif is_valid_integer(infos['month'],1):
                rep_dict['{MM}'] = '0'+infos['month']
            elif (infos['month'].lower() in valid_months.keys()):
                rep_dict['{MM}'] = month_to_number(infos['month'].lower())

    if '{DD}'in rep_dict.keys():
        rep_dict['{DD}'] = '00'
        if ('day' in infos):
            if is_valid_integer(infos['day'],2):
                rep_dict['{DD}'] = infos['day']
            elif is_valid_integer(infos['day'],1):
                rep_dict['{DD}'] = '0'+infos['day']

    if ('{J}' in rep_dict.keys()) or ('{Jabbr}' in rep_dict.keys()):
        if ('journal' in infos) and infos['journal']:
            rep_dict['{J}'] = validate_journal(infos['journal'])
            Jabbr = find_abbreviation_journal(infos['journal'])
            if Jabbr:
                rep_dict['{Jabbr}'] = Jabbr
            else:
                rep_dict['{Jabbr}'] = rep_dict['{J}']
        elif ('ejournal' in infos) and infos['ejournal']:
            rep_dict['{J}'] = infos['ejournal']
            rep_dict['{Jabbr}'] = infos['ejournal']
        else:
            rep_dict['{J}'] = '[NoJournal]'
            rep_dict['{Jabbr}'] = '[NoJourn]'

    author_info = ''
    if 'author' in infos.keys():
        author_info = infos['author']
    if 'authors' in infos.keys() and len(infos['authors'])>len(author_info):
        author_info = infos['authors']
    
    ListAuthorTags = ['{Aall}','{A3etal}','{Aetal}','{aAall}','{aA3etal}','{aAetal}']
    if any(item in rep_dict.keys() for item in ListAuthorTags): #Chec if any of the tag chosen by the user is one of the author tags defined in ListAuthorTags
        if author_info:
            # The variable author_info comes from the metadata genereated by pdf2bib, and its type/value depend on how the metadata was retrieved.
            # It will be either a string in the format "firstname1, secondname1 ...  lastname1 and firstname2, secondname2 ...  lastname2 etc."
            # or a list of dictionaries in the format  [{'given': 'Name1', 'family': 'LastName1'}, {'given': 'Name2', 'family': 'LastName2'}, ... [{'given': 'NameN', 'family': 'LastNameN'}]
            if isinstance(author_info,list):
                lastnames = [author['family'] for author in author_info if 'family' in author]
                firstnames = [author['given'] for author in author_info if 'given' in author]
            elif isinstance(author_info,str):
                authors = [author.strip() for author in author_info.split(" and ")]
                lastnames = [name.split()[-1] for name in authors]
                firstnames = [name.split()[:-1] if len(name.split())>1 else [''] for name in authors]   # The check on len(name.split())>1 is necessary to address the case in which                                                                                     #  the string name contains only one words (e.g. only the last name of the author is available)    
            
            if lastnames:                                                                                        
                rep_dict['{Aall}'] = ", ".join(lastnames)
                rep_dict['{A3etal}'] = ", ".join(lastnames[0:3])
                if len(lastnames)>3:
                    rep_dict['{A3etal}'] = rep_dict['{A3etal}'] + " et al."
                rep_dict['{Aetal}'] = lastnames[0] + (" et al." if len(lastnames)>1 else "")

                if firstnames: 
                    firstinitials = [firstname[0][0].upper()+"."  if len(firstname[0])>0 else "" for firstname in firstnames]
                    firstinitial_lastnames = [firstinitials + " " + lastname for (firstinitials,lastname) in zip(firstinitials,lastnames) ]
  
                    rep_dict['{aAall}'] = ", ".join(firstinitial_lastnames)
                    rep_dict['{aA3etal}'] = ", ".join(firstinitial_lastnames[0:3])
                    if len(firstinitial_lastnames)>3:
                        rep_dict['{aA3etal}'] = rep_dict['{aA3etal}'] + " et al."
                    rep_dict['{aAetal}'] = firstinitial_lastnames[0] + (" et al." if len(firstinitial_lastnames)>1 else "")
                else:
                    rep_dict['{aAall}'] = rep_dict['{Aall}']
                    rep_dict['{aA3etal}'] = rep_dict['{A3etal}']
                    rep_dict['{aAetal}'] = rep_dict['{Aetal}']
            else:
                for tag in ListAuthorTags:
                    rep_dict[tag] = '[NoAuthor]'

        else: #if author_info evalues to False, it means that it is either an empty string or an empty list, and thus the 'author' or 'authors' info were not present
              # in the bibtex metadata
            for tag in ListAuthorTags:
                rep_dict[tag] = '[NoAuthor]'

        #Check that none of the author strings is longer than max_length_authors. If they are, we truncate it
        for tag in ListAuthorTags:
            if tag in rep_dict.keys():
                rep_dict[tag] = rep_dict[tag][0:config.get('max_length_authors')]

    if (config.get('max_words_title') > 0 and 'title' in infos): #check if we need to limit the number of words in the title
        separator = " "
        n = config.get('max_words_title')
        title = infos['title']
        new_title = separator.join(title.split()[:n])
        infos['title'] = new_title

    if '{T}' in rep_dict.keys():
        if ('title' in infos) and infos['title']:
            rep_dict['{T}'] = infos['title']
        else:
            rep_dict['{T}'] = '[NoTitle]'

    #if '{Tcamel}' in rep_dict.keys():
    #    if ('title' in infos) and infos['title']:
    #        rep_dict['{Tcamel}'] = to_camel(infos['title'])
    #    else:
    #        rep_dict['{Tcamel}'] = '[NoTitle]'

    #if '{Tsnake}' in rep_dict.keys():
    #    if ('title' in infos) and infos['title']:
    #        rep_dict['{Tsnake}'] = to_snake(infos['title'])
    #    else:
    #        rep_dict['{Tsnake}'] = '[NoTitle]'
    
    #if '{Tkebab}' in rep_dict.keys():
    #    if ('title' in infos) and infos['title']:
    #        rep_dict['{Tkebab}'] = to_kebab(infos['title'])
    #    else:
    #        rep_dict['{Tkebab}'] = '[NoTitle]'


    if not ( config.get('case') == 'none' ):
        for key in rep_dict.keys():
            rep_dict[key] = string_transformations[config.get('case')](rep_dict[key])

    #Create the actual name
    for key in rep_dict.keys():
        format = format.replace(key, rep_dict[key])

    filename = sanitize(format)

    #Check that the filename string is not longer than max_length_filename, and truncate it in case. 
    filename = filename[0:config.get('max_length_filename')]
    return filename